@@ -12,7 +12,6 @@ from trytond.tools import grouped_slice
 
 __all__ = ['Location', 'Party', 'ProductsByLocationsStart',
     'ProductsByLocations']
-__metaclass__ = PoolMeta
 
 STATES = {
     'readonly': ~Eval('active'),
@@ -53,6 +52,8 @@ class Location(ModelSQL, ModelView):
     left = fields.Integer('Left', required=True, select=True)
     right = fields.Integer('Right', required=True, select=True)
     childs = fields.One2Many("stock.location", "parent", "Children")
+    warehouse = fields.Function(fields.Many2One('stock.location', 'Warehouse'),
+        'get_warehouse')
     input_location = fields.Many2One(
         "stock.location", "Input", states={
             'invisible': Eval('type') != 'warehouse',
@@ -159,9 +160,8 @@ class Location(ModelSQL, ModelView):
     def __register__(cls, module_name):
         TableHandler = backend.get('TableHandler')
         super(Location, cls).__register__(module_name)
-        cursor = Transaction().cursor
 
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.index_action(['left', 'right'], 'add')
 
     @classmethod
@@ -205,6 +205,15 @@ class Location(ModelSQL, ModelView):
     def check_xml_record(self, records, values):
         return True
 
+    def get_warehouse(self, name):
+        # Order by descending left to get the first one in the tree
+        locations = self.search([
+                ('parent', 'parent_of', [self.id]),
+                ('type', '=', 'warehouse'),
+                ], order=[('left', 'DESC')])
+        if locations:
+            return locations[0].id
+
     @classmethod
     def search_rec_name(cls, name, clause):
         locations = cls.search([
@@ -384,6 +393,7 @@ class Location(ModelSQL, ModelView):
 
 
 class Party:
+    __metaclass__ = PoolMeta
     __name__ = 'party.party'
     supplier_location = fields.Property(fields.Many2One('stock.location',
         'Supplier Location', domain=[('type', '=', 'supplier')],
