@@ -28,7 +28,6 @@ __all__ = ['ShipmentIn', 'ShipmentInReturn',
     'DeliveryNote', 'PickingList',
     'SupplierRestockingList', 'CustomerReturnRestockingList',
     'InteralShipmentReport']
-__metaclass__ = PoolMeta
 
 STATES = {
     'readonly': "state in ('cancel', 'done')",
@@ -38,7 +37,7 @@ STATES = {
 class ShipmentIn(Workflow, ModelSQL, ModelView):
     "Supplier Shipment"
     __name__ = 'stock.shipment.in'
-    _rec_name = 'code'
+    _rec_name = 'number'
     effective_date = fields.Date('Effective Date',
         states={
             'readonly': Eval('state').in_(['cancel', 'done']),
@@ -125,7 +124,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
         domain=[('company', '=', Eval('company'))], readonly=True,
         depends=['company'])
     origins = fields.Function(fields.Char('Origins'), 'get_origins')
-    code = fields.Char("Code", size=None, select=True, readonly=True)
+    number = fields.Char('Numer', size=None, select=True, readonly=True)
     state = fields.Selection([
         ('draft', 'Draft'),
         ('done', 'Done'),
@@ -170,7 +169,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         model_data = Table('ir_model_data')
         model = Table('ir_model')
         model_field = Table('ir_model_field')
@@ -207,9 +206,9 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
                 & (model_field.module == module_name)))
 
         old_table = 'stock_packing_in'
-        if TableHandler.table_exist(cursor, old_table):
-            TableHandler.table_rename(cursor, old_table, cls._table)
-        table = TableHandler(cursor, cls, module_name)
+        if TableHandler.table_exist(old_table):
+            TableHandler.table_rename(old_table, cls._table)
+        table = TableHandler(cls, module_name)
         for field in ('create_uid', 'write_uid', 'contact_address',
                 'warehouse', 'supplier'):
             table.drop_fk(field, table=old_table)
@@ -219,12 +218,16 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
         # Migration from 2.0:
         created_company = table.column_exist('company')
 
+        # Migration from 3.8: rename code into number
+        if table.column_exist('code'):
+            table.column_rename('code', 'number')
+
         super(ShipmentIn, cls).__register__(module_name)
 
         # Migration from 2.0:
         Move = Pool().get('stock.move')
         if (not created_company
-                and TableHandler.table_exist(cursor, Move._table)):
+                and TableHandler.table_exist(Move._table)):
             move = Move.__table__()
             cursor.execute(*sql_table.join(move,
                     condition=(Concat(cls.__name__ + ',', sql_table.id)
@@ -244,7 +247,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
             table.not_null_action('company', action='add')
 
         # Add index on create_date
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.index_action('create_date', action='add')
 
     @staticmethod
@@ -374,7 +377,8 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
         vlist = [x.copy() for x in vlist]
         config = Config(1)
         for values in vlist:
-            values['code'] = Sequence.get_id(config.shipment_in_sequence)
+            if values.get('number') is None:
+                values['number'] = Sequence.get_id(config.shipment_in_sequence)
         shipments = super(ShipmentIn, cls).create(vlist)
         cls._set_move_planned_date(shipments)
         return shipments
@@ -479,7 +483,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
 class ShipmentInReturn(Workflow, ModelSQL, ModelView):
     "Supplier Return Shipment"
     __name__ = 'stock.shipment.in.return'
-    _rec_name = 'code'
+    _rec_name = 'number'
     effective_date = fields.Date('Effective Date',
         states={
             'readonly': Eval('state').in_(['cancel', 'done']),
@@ -498,11 +502,26 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
                 Eval('context', {}).get('company', -1)),
             ],
         depends=['state'])
-    code = fields.Char("Code", size=None, select=True, readonly=True)
+    number = fields.Char('Number', size=None, select=True, readonly=True)
     reference = fields.Char("Reference", size=None, select=True,
         states={
             'readonly': Eval('state') != 'draft',
             }, depends=['state'])
+    supplier = fields.Many2One('party.party', 'Supplier',
+        states={
+            'readonly': (((Eval('state') != 'draft')
+                    | Eval('moves', [0]))
+                    & Eval('supplier', 0)),
+            }, required=True,
+        depends=['state', 'supplier'])
+    delivery_address = fields.Many2One('party.address', 'Delivery Address',
+        states={
+            'readonly': Eval('state') != 'draft',
+            },
+        domain=[
+            ('party', '=', Eval('supplier'))
+            ],
+        depends=['state', 'supplier'])
     from_location = fields.Many2One('stock.location', "From Location",
         required=True, states={
             'readonly': (Eval('state') != 'draft') | Eval('moves', [0]),
@@ -581,13 +600,13 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         sql_table = cls.__table__()
         # Migration from 1.2: packing renamed into shipment
         old_table = 'stock_packing_in_return'
-        if TableHandler.table_exist(cursor, old_table):
-            TableHandler.table_rename(cursor, old_table, cls._table)
-        table = TableHandler(cursor, cls, module_name)
+        if TableHandler.table_exist(old_table):
+            TableHandler.table_rename(old_table, cls._table)
+        table = TableHandler(cls, module_name)
         for field in ('create_uid', 'write_uid', 'from_location',
                 'to_location'):
             table.drop_fk(field, table=old_table)
@@ -597,12 +616,16 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
         # Migration from 2.0:
         created_company = table.column_exist('company')
 
+        # Migration from 3.8: rename code into number
+        if table.column_exist('code'):
+            table.column_rename('code', 'number')
+
         super(ShipmentInReturn, cls).__register__(module_name)
 
         # Migration from 2.0:
         Move = Pool().get('stock.move')
         if (not created_company
-                and TableHandler.table_exist(cursor, Move._table)):
+                and TableHandler.table_exist(Move._table)):
             move = Move.__table__()
             cursor.execute(*sql_table.join(move,
                     condition=(Concat(cls.__name__ + ',', sql_table.id)
@@ -622,7 +645,7 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
             table.not_null_action('company', action='add')
 
         # Add index on create_date
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.index_action('create_date', action='add')
 
     @staticmethod
@@ -633,6 +656,12 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
     def default_company():
         return Transaction().context.get('company')
 
+    @fields.depends('supplier')
+    def on_change_supplier(self):
+        if self.supplier:
+            self.delivery_address = self.supplier.address_get('delivery')
+            self.to_location = self.supplier.supplier_location
+
     @property
     def _move_planned_date(self):
         '''
@@ -666,8 +695,9 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
         vlist = [x.copy() for x in vlist]
         config = Config(1)
         for values in vlist:
-            values['code'] = Sequence.get_id(
-                config.shipment_in_return_sequence.id)
+            if values.get('number') is None:
+                values['number'] = Sequence.get_id(
+                    config.shipment_in_return_sequence.id)
         shipments = super(ShipmentInReturn, cls).create(vlist)
         cls._set_move_planned_date(shipments)
         return shipments
@@ -757,7 +787,7 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
 class ShipmentOut(Workflow, ModelSQL, ModelView):
     "Customer Shipment"
     __name__ = 'stock.shipment.out'
-    _rec_name = 'code'
+    _rec_name = 'number'
     effective_date = fields.Date('Effective Date',
         states={
             'readonly': Eval('state').in_(['cancel', 'done']),
@@ -837,7 +867,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         domain=[('company', '=', Eval('company'))], depends=['company'],
         readonly=True)
     origins = fields.Function(fields.Char('Origins'), 'get_origins')
-    code = fields.Char("Code", size=None, select=True, readonly=True)
+    number = fields.Char('Number', size=None, select=True, readonly=True)
     state = fields.Selection([
         ('draft', 'Draft'),
         ('done', 'Done'),
@@ -905,14 +935,14 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         sql_table = cls.__table__()
         # Migration from 1.2: packing renamed into shipment
         old_table = 'stock_packing_out'
-        if TableHandler.table_exist(cursor, old_table):
-            TableHandler.table_rename(cursor, old_table, cls._table)
+        if TableHandler.table_exist(old_table):
+            TableHandler.table_rename(old_table, cls._table)
 
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         for field in ('create_uid', 'write_uid', 'delivery_address',
                 'warehouse', 'customer'):
             table.drop_fk(field, table=old_table)
@@ -922,12 +952,16 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         # Migration from 2.0:
         created_company = table.column_exist('company')
 
+        # Migration from 3.8: rename code into number
+        if table.column_exist('code'):
+            table.column_rename('code', 'number')
+
         super(ShipmentOut, cls).__register__(module_name)
 
         # Migration from 2.0:
         Move = Pool().get('stock.move')
         if (not created_company
-                and TableHandler.table_exist(cursor, Move._table)):
+                and TableHandler.table_exist(Move._table)):
             move = Move.__table__()
             cursor.execute(*sql_table.join(move,
                     condition=(Concat(cls.__name__ + ',', sql_table.id)
@@ -947,7 +981,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
             table.not_null_action('company', action='add')
 
         # Migration from 1.0 customer_location is no more used
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.drop_column('customer_location', exception=True)
 
         # Add index on create_date
@@ -1217,18 +1251,26 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         Set planned date of moves for the shipments
         '''
         Move = Pool().get('stock.move')
+        to_write = []
         for shipment in shipments:
             outgoing_date, inventory_date = shipment._move_planned_date
-            Move.write([x for x in shipment.outgoing_moves
+            out_moves_to_write = [x for x in shipment.outgoing_moves
                     if (x.state not in ('assigned', 'done', 'cancel')
-                        and x.planned_date != outgoing_date)], {
-                    'planned_date': outgoing_date,
-                    })
-            Move.write([x for x in shipment.inventory_moves
+                        and x.planned_date != outgoing_date)]
+            if out_moves_to_write:
+                to_write.extend((out_moves_to_write, {
+                        'planned_date': outgoing_date,
+                        }))
+
+            inv_moves_to_write = [x for x in shipment.inventory_moves
                     if (x.state not in ('assigned', 'done', 'cancel')
-                        and x.planned_date != inventory_date)], {
-                    'planned_date': inventory_date,
-                    })
+                        and x.planned_date != inventory_date)]
+            if inv_moves_to_write:
+                to_write.extend((inv_moves_to_write, {
+                        'planned_date': inventory_date,
+                        }))
+        if to_write:
+            Move.write(*to_write)
 
     @classmethod
     def create(cls, vlist):
@@ -1239,7 +1281,9 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         vlist = [x.copy() for x in vlist]
         config = Config(1)
         for values in vlist:
-            values['code'] = Sequence.get_id(config.shipment_out_sequence.id)
+            if values.get('number') is None:
+                values['number'] = Sequence.get_id(
+                    config.shipment_out_sequence.id)
         shipments = super(ShipmentOut, cls).create(vlist)
         cls._set_move_planned_date(shipments)
         return shipments
@@ -1309,7 +1353,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
 class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
     "Customer Return Shipment"
     __name__ = 'stock.shipment.out.return'
-    _rec_name = 'code'
+    _rec_name = 'number'
     effective_date = fields.Date('Effective Date',
         states={
             'readonly': Eval('state').in_(['cancel', 'done']),
@@ -1388,7 +1432,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
         domain=[('company', '=', Eval('company'))], depends=['company'],
         readonly=True)
     origins = fields.Function(fields.Char('Origins'), 'get_origins')
-    code = fields.Char("Code", size=None, select=True, readonly=True)
+    number = fields.Char('Number', size=None, select=True, readonly=True)
     state = fields.Selection([
         ('draft', 'Draft'),
         ('done', 'Done'),
@@ -1430,14 +1474,14 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         sql_table = cls.__table__()
         # Migration from 1.2: packing renamed into shipment
         old_table = 'stock_packing_out_return'
-        if TableHandler.table_exist(cursor, old_table):
-            TableHandler.table_rename(cursor, old_table, cls._table)
+        if TableHandler.table_exist(old_table):
+            TableHandler.table_rename(old_table, cls._table)
 
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         for field in ('create_uid', 'write_uid', 'delivery_address',
                 'warehouse', 'customer'):
             table.drop_fk(field, table=old_table)
@@ -1447,12 +1491,16 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
         # Migration from 2.0:
         created_company = table.column_exist('company')
 
+        # Migration from 3.8: rename code into number
+        if table.column_exist('code'):
+            table.column_rename('code', 'number')
+
         super(ShipmentOutReturn, cls).__register__(module_name)
 
         # Migration from 2.0:
         Move = Pool().get('stock.move')
         if (not created_company
-                and TableHandler.table_exist(cursor, Move._table)):
+                and TableHandler.table_exist(Move._table)):
             move = Move.__table__()
             cursor.execute(*sql_table.join(move,
                     condition=(Concat(cls.__name__ + ',', sql_table.id)
@@ -1472,7 +1520,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
             table.not_null_action('company', action='add')
 
         # Add index on create_date
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.index_action('create_date', action='add')
 
     @staticmethod
@@ -1592,8 +1640,9 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
         vlist = [x.copy() for x in vlist]
         config = Config(1)
         for values in vlist:
-            values['code'] = Sequence.get_id(
-                    config.shipment_out_return_sequence.id)
+            if values.get('number') is None:
+                values['number'] = Sequence.get_id(
+                        config.shipment_out_return_sequence.id)
         shipments = super(ShipmentOutReturn, cls).create(vlist)
         cls._set_move_planned_date(shipments)
         return shipments
@@ -1746,7 +1795,7 @@ class AssignShipmentOut(Wizard):
 class ShipmentInternal(Workflow, ModelSQL, ModelView):
     "Internal Shipment"
     __name__ = 'stock.shipment.internal'
-    _rec_name = 'code'
+    _rec_name = 'number'
     effective_date = fields.Date('Effective Date',
         states={
             'readonly': Eval('state').in_(['cancel', 'done']),
@@ -1765,7 +1814,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
                 Eval('context', {}).get('company', -1)),
             ],
         depends=['state'])
-    code = fields.Char("Code", size=None, select=True, readonly=True)
+    number = fields.Char('Number', size=None, select=True, readonly=True)
     reference = fields.Char("Reference", size=None, select=True,
         states={
             'readonly': Eval('state') != 'draft',
@@ -1864,13 +1913,13 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         sql_table = cls.__table__()
         # Migration from 1.2: packing renamed into shipment
         old_table = 'stock_packing_internal'
-        if TableHandler.table_exist(cursor, old_table):
-            TableHandler.table_rename(cursor, old_table, cls._table)
-        table = TableHandler(cursor, cls, module_name)
+        if TableHandler.table_exist(old_table):
+            TableHandler.table_rename(old_table, cls._table)
+        table = TableHandler(cls, module_name)
         for field in ('create_uid', 'write_uid', 'from_location',
                 'to_location'):
             table.drop_fk(field, table=old_table)
@@ -1880,12 +1929,16 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
         # Migration from 2.0:
         created_company = table.column_exist('company')
 
+        # Migration from 3.8:
+        if table.column_exist('code'):
+            table.column_rename('code', 'number')
+
         super(ShipmentInternal, cls).__register__(module_name)
 
         # Migration from 2.0:
         Move = Pool().get('stock.move')
         if (not created_company
-                and TableHandler.table_exist(cursor, Move._table)):
+                and TableHandler.table_exist(Move._table)):
             move = Move.__table__()
             cursor.execute(*sql_table.join(move,
                     condition=(Concat(cls.__name__ + ',', sql_table.id)
@@ -1905,7 +1958,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
             table.not_null_action('company', action='add')
 
         # Add index on create_date
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.index_action('create_date', action='add')
 
     @staticmethod
@@ -1925,8 +1978,9 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
         vlist = [x.copy() for x in vlist]
         config = Config(1)
         for values in vlist:
-            values['code'] = Sequence.get_id(
-                    config.shipment_internal_sequence.id)
+            if values.get('number') is None:
+                values['number'] = Sequence.get_id(
+                        config.shipment_internal_sequence.id)
         return super(ShipmentInternal, cls).create(vlist)
 
     @classmethod
@@ -2020,6 +2074,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
 
 
 class Address:
+    __metaclass__ = PoolMeta
     __name__ = 'party.address'
     delivery = fields.Boolean('Delivery')
 
